@@ -4,7 +4,7 @@
 #
 
 package Parallel::WorkUnit;
-$Parallel::WorkUnit::VERSION = '1.113'; # TRIAL
+$Parallel::WorkUnit::VERSION = '1.114'; # TRIAL
 use v5.8;
 
 # ABSTRACT: Provide multi-paradigm forking with ability to pass back data
@@ -47,6 +47,11 @@ use Storable;
 use namespace::autoclean;
 
 
+my @ALL_WU;    # Holds all active work units so child processes can't
+               # mess with parent work units
+               # Note it holds a reference (strong) to a reference
+               # (weak).
+
 subtype 'Parallel::WorkUnit::PositiveInt', as 'Int',
   where { $_ > 0 },
   message { "The number you provided, $_, was not a positive number" };
@@ -105,6 +110,12 @@ has '_queue' => (
     is       => 'rw',
     init_arg => undef
 );
+# This only gets used on Win32.
+#
+has '_child_queue' => (
+    is       => 'rw',
+    init_arg => undef
+);
 
 # This only gets used on Win32
 has '_count' => (
@@ -114,13 +125,6 @@ has '_count' => (
     default  => 1
 );
 
-has '_parent_pid' => (
-    is       => 'rw',
-    isa      => 'Int',
-    init_arg => undef,
-    default  => $$,
-);
-
 # Children queued
 has '_queued_children' => (
     is       => 'rw',
@@ -139,6 +143,13 @@ sub BUILD {
         $self->_queue( Thread::Queue->new() );
     }
 
+    # Make a weak reference and shove it into the ALL_WU array
+    weaken $self;
+    push @ALL_WU, \$self;
+
+    # Do some housekeeping on @ALL_WU, so it is somewhat bounded
+    @ALL_WU = grep { defined $$_ } @ALL_WU;
+
     return;
 }
 
@@ -149,7 +160,7 @@ sub async {
     my $sub  = shift;
 
     # Test $sub to make sure it is a code ref or a sub ref
-    if (! _codelike($sub)) {
+    if ( !_codelike($sub) ) {
         croak("Parameter to async() is not a code (or codelike) reference");
     }
 
@@ -164,7 +175,7 @@ sub async {
         my $selfref = $self;
         weaken $selfref;
         $callback = sub {
-            if (defined $selfref) { # Incase this went away
+            if ( defined $selfref ) {    # Incase this went away
                 @{ $selfref->_ordered_responses }[$cbnum] = shift;
             }
         };
@@ -232,6 +243,12 @@ sub _child {
     if ( scalar(@_) != 4 ) { confess 'invalid call'; }
     my ( $self, $sub, $pipe, $pid ) = @_;
 
+    # Cleanup ALL_WU
+    @ALL_WU = grep { defined $$_ } @ALL_WU;
+    foreach my $wu ( map { $$_ } @ALL_WU ) {
+        $wu->_clear_all( $wu == $self );
+    }
+
     # We are in the child process
     if ($use_anyevent_pipe) {
         $pipe = $pipe->[1];
@@ -438,7 +455,7 @@ sub queue {
     my $sub  = shift;
 
     # Test $sub to make sure it is a code ref or a sub ref
-    if (! _codelike($sub)) {
+    if ( !_codelike($sub) ) {
         croak("Parameter to queue() is not a code (or codelike) reference");
     }
 
@@ -484,7 +501,7 @@ sub _send {
     }
 
     if ($use_thread_queue) {
-        $self->_queue()->enqueue($pid);
+        $self->_child_queue()->enqueue($pid);
     }
 
     $fh->write($type);
@@ -651,6 +668,35 @@ sub _add_anyevent_watcher {
     return;
 }
 
+# Used to clear all sub-processes, etc, in child process.
+#
+# Parameter one should be the object to clear
+# Parameter 2 determines whether or not we keep _queue (saving into
+# _child_queue)
+sub _clear_all {
+    if ( $#_ != 1 ) { confess 'invalid call' }
+    my ( $self, $keep_queue ) = @_;
+
+    $self->_cv(undef);
+    $self->_last_error(undef);
+    $self->_ordered_count(0);
+    $self->_ordered_responses( [] );
+    $self->_subprocs( {} );
+    $self->_count(1);
+    $self->_queued_children( [] );
+
+    if ( defined( $self->_queue() ) ) {
+        if ($keep_queue) {
+            $self->_child_queue( $self->_queue() );
+        } else {
+            $self->_child_queue(undef);
+        }
+        $self->_queue( Thread::Queue->new() );
+    }
+
+    return;
+}
+
 # Tests to see if something is codelike
 #
 # Borrowed from Params::Util (written by Adam Kennedy)
@@ -658,8 +704,8 @@ sub _codelike {
     if ( scalar(@_) != 1 ) { confess 'invalid call' }
     my $thing = shift;
 
-    if (reftype($thing)) { return 1; }
-    if (blessed($thing) && overload::Method($thing, '()')) { return 1; }
+    if ( reftype($thing) ) { return 1; }
+    if ( blessed($thing) && overload::Method( $thing, '()' ) ) { return 1; }
 
     return;
 }
@@ -673,12 +719,6 @@ sub DEMOLISH {
     # different process IDs for this warning.
     if ($use_threads) { return; }
 
-    # When we fork(), we might have other work unit processes hanging
-    # around
-    # XXX There should be a better way of handling this. We don't want
-    # pipes and the like held open by multiple processes.
-    if ( $self->_parent_pid != $$ ) { return; }
-
     if ( scalar( keys %{ $self->_subprocs } ) ) {
         warn "Warning: Subprocesses running when Parallel::WorkUnit object destroyed\n";
     }
@@ -702,7 +742,7 @@ Parallel::WorkUnit - Provide multi-paradigm forking with ability to pass back da
 
 =head1 VERSION
 
-version 1.113
+version 1.114
 
 =head1 SYNOPSIS
 
